@@ -51,6 +51,7 @@ DISCNUM_FID        = "TPOS";
 USERTEXT_FID       = "TXXX";
 CDID_FID           = "MCDI";
 IMAGE_FID          = "APIC";
+OBJECT_FID         = "GEOB";
 URL_COMMERCIAL_FID = "WCOM";
 URL_COPYRIGHT_FID  = "WCOP";
 URL_AUDIOFILE_FID  = "WOAF";
@@ -185,7 +186,7 @@ frameDesc = { "AENC": "Audio encryption",
 
 
 # mapping of 2.2 frames to 2.3/2.4
-TAGS2_2_TO_TAGS_2_3_AND_4 = {                  
+TAGS2_2_TO_TAGS_2_3_AND_4 = {
     "TT1" : "TIT1", # CONTENTGROUP content group description
     "TT2" : "TIT2", # TITLE title/songname/content description
     "TT3" : "TIT3", # SUBTITLE subtitle/description refinement
@@ -265,6 +266,7 @@ COMMENT_FRAME_RX = re.compile("^" + COMMENT_FID + "$");
 LYRICS_FRAME_RX = re.compile("^" + LYRICS_FID + "$");
 CDID_FRAME_RX = re.compile("^" + CDID_FID + "$");
 IMAGE_FRAME_RX = re.compile("^" + IMAGE_FID + "$");
+OBJECT_FRAME_RX = re.compile("^" + OBJECT_FID + "$");
 PLAYCOUNT_FRAME_RX = re.compile("^" + PLAYCOUNT_FID + "$");
 UNIQUE_FILE_ID_FRAME_RX = re.compile("^" + UNIQUE_FILE_ID_FID + "$");
 
@@ -341,14 +343,14 @@ class FrameHeader:
 
    # 2.4 not only added flag bits, but also reordered the previously defined
    # flags.  So these are mapped once we know the version.
-   TAG_ALTER   = None;   
-   FILE_ALTER  = None;   
-   READ_ONLY   = None;   
-   COMPRESSION = None;   
-   ENCRYPTION  = None;   
-   GROUPING    = None;   
-   UNSYNC      = None;   
-   DATA_LEN    = None;   
+   TAG_ALTER   = None;
+   FILE_ALTER  = None;
+   READ_ONLY   = None;
+   COMPRESSION = None;
+   ENCRYPTION  = None;
+   GROUPING    = None;
+   UNSYNC      = None;
+   DATA_LEN    = None;
 
    # Constructor.
    def __init__(self, tagHeader = None):
@@ -1445,6 +1447,151 @@ class ImageFrame(Frame):
          raise FrameException("Invalid APIC picture type: %d" % t);
    picTypeToString = staticmethod(picTypeToString);
 
+################################################################################
+# This class refers to the GEOB frame
+class ObjectFrame(Frame):
+   mimeType = None;
+   description = u"";
+   filename = u"";
+   objectData = None;
+
+   def __init__(self, frameHeader, data = None,
+                desc = u"", filename = u"",
+                objectData = None, mimeType = None,
+                encoding = DEFAULT_ENCODING):
+       Frame.__init__(self, frameHeader);
+       if data != None:
+           self._set(data, frameHeader);
+       else:
+           assert(isinstance(desc, unicode));
+           self.description = desc;
+           assert(isinstance(filename, unicode));
+           self.filename = filename;
+           self.encoding = encoding;
+           assert(mimeType);
+           self.mimeType = mimeType;
+           assert(objectData);
+           self.objectData = objectData;
+
+   # Factory method
+   def create(objFile, mime = u"", desc = u"", filename = None,
+              encoding = DEFAULT_ENCODING):
+       if filename == None:
+           filename = unicode(os.path.basename(objFile));
+       if not isinstance(desc, unicode) or \
+          (not isinstance(filename, unicode) and filename != ""):
+           raise FrameException("Wrong description and/or filename type.");
+       # Load file
+       fp = file(objFile, "rb");
+       objData = fp.read();
+       if mime:
+           print("Using specified mime type %s" % mime);
+       else:
+           mt = mimetypes.guess_type(objFile);
+           if not mt[0]:
+               raise FrameException("Unable to guess mime-type for %s" %
+                                    objFile);
+           mime = mt[0];
+           print("Guessing mime type %s" % mime);
+
+       frameData = DEFAULT_ENCODING;
+       frameData += mime + "\x00";
+       frameData += filename.encode(id3EncodingToString(encoding)) + "\x00";
+       frameData += desc.encode(id3EncodingToString(encoding)) + "\x00";
+       frameData += objData;
+
+       frameHeader = FrameHeader();
+       frameHeader.id = OBJECT_FID;
+       return ObjectFrame(frameHeader, data = frameData);
+   # Make create a static method.  Odd....
+   create = staticmethod(create);
+
+   # Data string format:
+   # <Header for 'General encapsulated object', ID: "GEOB">
+   #  Text encoding          $xx
+   #  MIME type              <text string> $00
+   #  Filename               <text string according to encoding> $00 (00)
+   #  Content description    <text string according to encoding> $00 (00)
+   #  Encapsulated object    <binary data>
+   def _set(self, data, frameHeader = None):
+      assert(frameHeader);
+      if not OBJECT_FRAME_RX.match(frameHeader.id):
+         raise FrameException("Invalid frame id for ObjectFrame: " +\
+                              frameHeader.id);
+
+      data = self.disassembleFrame(data);
+
+      input = StringIO(data);
+      TRACE_MSG("GEOB frame data size: " + str(len(data)));
+      self.encoding = input.read(1);
+      TRACE_MSG("GEOB encoding: " + id3EncodingToString(self.encoding));
+
+      # Mime type
+      self.mimeType = "";
+      if self.header.minorVersion != 2:
+          ch = input.read(1);
+          while ch != "\x00":
+              self.mimeType += ch;
+              ch = input.read(1);
+      else:
+          # v2.2 (OBSOLETE) special case
+          self.mimeType = input.read(3);
+      TRACE_MSG("GEOB mime type: " + self.mimeType);
+      if strictID3() and not self.mimeType:
+         raise FrameException("GEOB frame does not contain a mime type");
+      if strictID3() and self.mimeType.find("/") == -1:
+         raise FrameException("GEOB frame does not contain a valid mime type");
+
+      self.filename = u"";
+      self.description = u"";
+
+      # Remaining data is a NULL separated filename, description and object data
+      buffer = input.read();
+      input.close();
+
+      (filename, buffer) = splitUnicode(buffer, self.encoding);
+      (desc, obj) = splitUnicode(buffer, self.encoding);
+      TRACE_MSG("filename len: %d" % len(filename));
+      TRACE_MSG("description len: %d" % len(desc));
+      TRACE_MSG("data len: %d" % len(obj));
+      self.filename = unicode(filename, id3EncodingToString(self.encoding));
+      self.description = unicode(desc, id3EncodingToString(self.encoding));
+      TRACE_MSG("GEOB filename: " + self.filename);
+      TRACE_MSG("GEOB description: " + self.description);
+
+      self.objectData = obj;
+      TRACE_MSG("GEOB data: " + str(len(self.objectData)) + " bytes");
+      if strictID3() and not self.objectData:
+         raise FrameException("GEOB frame does not contain any data");
+
+
+   def writeFile(self, path = "./", name = None):
+      if not self.objectData:
+         raise IOError("Fetching remote object files is not implemented.");
+      if not name:
+         name = self.getDefaultFileName();
+      objectFile = os.path.join(path, name);
+
+      f = file(objectFile, "wb");
+      f.write(self.objectData);
+      f.flush();
+      f.close();
+   def getDefaultFileName(self, suffix = ""):
+      nameStr = self.filename;
+      if suffix:
+          nameStr += suffix;
+      nameStr = nameStr +  "." + self.mimeType.split("/")[1];
+      return nameStr;
+
+   def render(self):
+      data = self.encoding + self.mimeType + "\x00" +\
+             self.filename.encode(id3EncodingToString(self.encoding)) +\
+             self.getTextDelim() +\
+             self.description.encode(id3EncodingToString(self.encoding)) +\
+             self.getTextDelim() +\
+             self.objectData;
+      return self.assembleFrame(data);
+
 class PlayCountFrame(Frame):
     count = None;
 
@@ -1459,7 +1606,7 @@ class PlayCountFrame(Frame):
         assert(frameHeader);
         assert(len(data) >= 4);
         self.count = long(bytes2dec(data));
-        
+
     def render(self):
         data = dec2bytes(self.count, 32);
         return self.assembleFrame(data);
@@ -1489,7 +1636,7 @@ class UniqueFileIDFrame(Frame):
         if strictID3() and (len(self.owner_id) == 0 or
                             len(self.id) == 0 or len(self.id) > 64):
             raise FrameException("Invalid UFID frame");
-        
+
     def render(self):
         data = self.owner_id + "\x00" + self.id;
         return self.assembleFrame(data);
@@ -1614,7 +1761,7 @@ class FrameSet(list):
       for f in self:
          sz += len(f.render());
       return sz;
-   
+
    def setTagHeader(self, tagHeader):
       self.tagHeader = tagHeader;
       for f in self:
@@ -1683,6 +1830,15 @@ class FrameSet(list):
                raise FrameException("Multiple %s frames with the same "\
                                     "content descriptor not allowed." % fid);
 
+      # Object (GEOB) frame restrictions.
+      # Multiples must have a unique content desciptor.
+      if OBJECT_FRAME_RX.match(fid) and self[fid] and strictID3():
+         objectFrames = self[fid];
+         for frm in objectFrames:
+            if frm.description == frame.description:
+               raise FrameException("Multiple %s frames with the same "\
+                                    "content descriptor not allowed." % fid);
+
       # Play count frame (PCNT).  There may be only one
       if PLAYCOUNT_FRAME_RX.match(fid) and self[fid]:
          raise FrameException("Multiple %s frames not allowed." % fid);
@@ -1843,7 +1999,7 @@ class FrameSet(list):
           self.addFrame(UserTextFrame(h, encoding = encoding,
                                       description = description,
                                       text = txt));
-         
+
    # This method removes all frames with the matching frame ID.
    # The number of frames removed is returned.
    # Note that calling this method with a key like "COMM" may remove more
@@ -1945,6 +2101,9 @@ def createFrame(frameHeader, data):
   # Attached picture
   elif IMAGE_FRAME_RX.match(frameHeader.id):
      f = ImageFrame(frameHeader, data = data);
+  # Encapsulated object
+  elif OBJECT_FRAME_RX.match(frameHeader.id):
+     f = ObjectFrame(frameHeader, data = data);
   # Play count
   elif PLAYCOUNT_FRAME_RX.match(frameHeader.id):
      f = PlayCountFrame(frameHeader, data = data);
@@ -1962,6 +2121,4 @@ def map2_2FrameId(originalId):
     if not TAGS2_2_TO_TAGS_2_3_AND_4.has_key(originalId):
         return originalId
     return TAGS2_2_TO_TAGS_2_3_AND_4[originalId]
-        
-        
-    
+
