@@ -119,17 +119,22 @@ class Header:
    # This may throw an Mp3Exception if the header is malformed.
    def decode(self, header):
       # MPEG audio version from bits 19 and 20.
-      if not header & (1 << 20) and header & (1 << 19):
-         raise Mp3Exception("Illegal MPEG audio version");
-      elif not header & (1 << 20) and not header & (1 << 19):
-         self.version = 2.5;
+      bit20 = header & (1 << 20)
+      bit19 = header & (1 << 19)
+      if not bit20 and bit19:
+          # http://www.dv.co.yu/mpgscript/mpeghdr.htm states that this mpeg
+          # version is "reserved", yet I see more and more mp3s using it.
+          # FIXME: I can't locate what version this is supposed to be, so use
+          #        v2.5.
+          self.version = 2.5;
+      elif not bit20 and not bit19:
+          self.version = 2.5;
       else:
-         if not header & (1 << 19):
-            self.version = 2.0;
-         else:
-            self.version = 1.0;
+          if not bit19:
+              self.version = 2.0;
+          else:
+              self.version = 1.0;
 
-      
       # MPEG audio layer from bits 18 and 17.
       if not header & (1 << 18) and not header & (1 << 17):
          raise Mp3Exception("Illegal MPEG layer value");
