@@ -1,6 +1,6 @@
 ###############################################################################
 #
-#  Copyright (C) 2002-2005  Travis Shirk <travis@pobox.com>
+#  Copyright (C) 2002-2006  Travis Shirk <travis@pobox.com>
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -25,6 +25,7 @@ import eyeD3.utils;
 import eyeD3.mp3;
 from frames import *;
 from binfuncs import *;
+import math;
 
 ID3_V1_COMMENT_DESC = "ID3 v1 Comment";
 
@@ -50,7 +51,7 @@ class TagHeader:
 
    # The size in the most recently parsed header.
    tagSize = 0;
-   
+
    # Constructor
    def __init__(self):
       self.clear();
@@ -168,11 +169,11 @@ class ExtendedTagHeader:
    restrictions = 0;
 
    def isUpdate(self):
-      return self.flags & 0x40;
+       return self.flags & 0x40;
    def hasCRC(self):
-      return self.flags & 0x20;
-   def hasRestrictions(self):
-      return self.flags & 0x10;
+       return self.flags & 0x20;
+   def hasRestrictions(self, minor_version = None):
+       return self.flags & 0x10;
 
    def setSizeRestrictions(self, v):
       assert(v >= 0 and v <= 3);
@@ -246,41 +247,61 @@ class ExtendedTagHeader:
          return "All images are exactly 64x64 pixels, unless required "\
                 "otherwise.";
 
+   def _syncsafeCRC(self):
+       bites = ""
+       bites += chr((self.crc >> 28) & 0x7f);
+       bites += chr((self.crc >> 21) & 0x7f);
+       bites += chr((self.crc >> 14) & 0x7f);
+       bites += chr((self.crc >>  7) & 0x7f);
+       bites += chr((self.crc >>  0) & 0x7f);
+       return bites;
+
+
    def render(self, header, frameData, padding = 0):
       assert(header.majorVersion == 2);
 
       data = "";
+      crc = None;
       if header.minorVersion == 4:
          # Version 2.4
-         crc = None;
          size = 6;
          # Extended flags.
          if self.isUpdate():
-            data = "\x00";
+            data += "\x00";
          if self.hasCRC():
             data += "\x05";
-            self.crc = binascii.crc32(data);
-            data += bin2bytes(dec2bin(self.crc, 40));
+            # XXX: Using the absolute value of the CRC.  The spec is unclear
+            # about the type of this data.
+            self.crc = int(math.fabs(binascii.crc32(frameData +\
+                                                    ("\x00" * padding))));
+            crc_data = self._syncsafeCRC();
+            if len(crc_data) < 5:
+                crc_data = ("\x00" * (5 - len(crc_data))) + crc_data
+            assert(len(crc_data) == 5)
+            data += crc_data
          if self.hasRestrictions():
             data += "\x01";
             assert(len(self.restrictions) == 1);
             data += self.restrictions;
+         TRACE_MSG("Rendered extended header data (%d bytes)" % len(data));
 
          # Extended header size.
          size = bin2bytes(bin2synchsafe(dec2bin(len(data) + 6, 32)))
          assert(len(size) == 4);
 
-         assert(len(self.flags) == 2);
-         data = size + "\x01" + self.flags + data;
+         data = size + "\x01" + bin2bytes(dec2bin(self.flags)) + data;
+         TRACE_MSG("Rendered extended header of size %d" % len(data));
       else:
          # Version 2.3
-         crc = None;
-         size = 6;
+         size = 10;
          # Extended flags.
          f = [0] * 16;
          if self.hasCRC():
             f[0] = 1;
-            self.crc = binascii.crc32(data);
+            # XXX: Using the absolute value of the CRC.  The spec is unclear
+            # about the type of this type.
+            self.crc = int(math.fabs(binascii.crc32(frameData +\
+                                                    ("\x00" * padding))));
             crc = bin2bytes(dec2bin(self.crc));
             assert(len(crc) == 4);
             size += 4;
@@ -295,19 +316,20 @@ class ExtendedTagHeader:
          data = size + flags + paddingSize;
          if crc:
             data += crc;
-      return data; 
+      return data;
 
    # Only call this when you *know* there is an extened header.
    def parse(self, fp, header):
       assert(header.majorVersion == 2);
 
+      TRACE_MSG("Parsing extended header @ 0x%x" % fp.tell());
       # First 4 bytes is the size of the extended header.
       data = fp.read(4);
       if header.minorVersion == 4:
-         TRACE_MSG("Parsing extended header for v2.4");
          # sync-safe
          sz = bin2dec(bytes2bin(data, 7));
-         TRACE_MSG("Extended header size: %d" % (sz - 4));
+         TRACE_MSG("Extended header size (not including 4 byte size): %d" %\
+                   (sz - 4));
          data = fp.read(sz - 4);
 
          if ord(data[0]) != 1 or (ord(data[1]) & 0x8f):
@@ -339,12 +361,15 @@ class ExtendedTagHeader:
             self.restrictions = ord(data[offset]);
             offset += 1;
       else:
-         TRACE_MSG("Parsing extended header for v2.3");
          # v2.3 is totally different... *sigh*
          sz = bin2dec(bytes2bin(data));
-         TRACE_MSG("Extended header size: %d" % sz);
-         data = fp.read(sz);
+         TRACE_MSG("Extended header size (not including 4 bytes size): %d" %\
+                   sz);
          tmpFlags = fp.read(2);
+         # Read the padding size, but it'll be computed during the parse.
+         ps = fp.read(4);
+         TRACE_MSG("Extended header says there is %d bytes of padding" %\
+                   bin2dec(bytes2bin(ps)));
          # Make this look like a v2.4 mask.
          self.flags = ord(tmpFlags[0]) >> 2;
          if self.hasCRC():
@@ -352,8 +377,6 @@ class ExtendedTagHeader:
             crcData = fp.read(4);
             self.crc = bin2dec(bytes2bin(crcData));
             TRACE_MSG("Extended header CRC: %d" % self.crc);
-         # Read the padding size, but it'll be computed during the parse.
-         fp.read(4);
 
 
 ################################################################################
@@ -392,7 +415,6 @@ class Tag:
       self.header = TagHeader();
       self.frames = FrameSet(self.header);
       self.iterIndex = None;
-      self.linkedFile = None;
 
    # Returns an read-only iterator for all frames.
    def __iter__(self):
@@ -421,15 +443,16 @@ class Tag:
    # Converts all ID3v1 data into ID3v2 frames internally.
    # May throw IOError, or TagException if parsing fails.
    def link(self, f, v = ID3_ANY_VERSION):
+      self.linkedFile = None;
       self.clear();
 
       fileName = "";
       if isinstance(f, file):
          fileName = f.name;
-      elif isinstance(f, str):
+      elif isinstance(f, str) or isinstance(f, unicode):
          fileName = f;
       else:
-         raise TagException("Invalid type passed to Tag.link: " + 
+         raise TagException("Invalid type passed to Tag.link: " +
                             str(type(f)));
 
       if v != ID3_V1 and v != ID3_V2 and v != ID3_ANY_VERSION:
@@ -484,7 +507,7 @@ class Tag:
          self.getTrackNum()[0] != None and version != ID3_V1_0:
          version = ID3_V1_1;
          self.setVersion(version);
-      
+
       # If there are no frames then simply remove the current tag.
       if len(self.frames) == 0:
          self.remove(version);
@@ -516,7 +539,7 @@ class Tag:
       if version == ID3_CURRENT_VERSION:
          version = self.getVersion();
 
-      retval = 0;   
+      retval = 0;
       if version & ID3_V1 or version == ID3_ANY_VERSION:
          tagFile = file(self.linkedFile.name, "r+b");
          tagFile.seek(-128, 2);
@@ -531,20 +554,31 @@ class Tag:
           self.header.tagSize:
          tagFile = file(self.linkedFile.name, "r+b");
          if tagFile.read(3) == "ID3":
+            # FIXME:
+            print "POINT #2"
             TRACE_MSG("Removing ID3 v2.x Tag");
             tagSize = self.header.tagSize + self.header.SIZE;
             tagFile.seek(tagSize);
-            data = tagFile.read();
-            tagFile.seek(0);
-            tagFile.write(data);
+
+            # Open tmp file
+            tmpName = tempfile.mktemp();
+            tmpFile = file(tmpName, "w+b");
+
+            # Write audio data in chunks
+            self.__copyRemaining(tagFile, tmpFile);
             tagFile.truncate();
             tagFile.close();
+
+            tmpFile.close();
+
+            # Move tmp to orig.
+            shutil.copyfile(tmpName, self.linkedFile.name);
+            os.unlink(tmpName);
+
             retval |= 1;
 
       return retval;
 
-
-
    # Get artist.  There are a few frames that can contain this information,
    # and they are subtley different. 
    #   eyeD3.frames.ARTIST_FID - Lead performer(s)/Soloist(s)
@@ -568,8 +602,8 @@ class Tag:
          f = self.frames[fid];
          if f:
              return f[0].text;
-      return u""; 
- 
+      return u"";
+
    def getAlbum(self):
       f = self.frames[ALBUM_FID];
       if f:
@@ -604,7 +638,7 @@ class Tag:
            return dateFrame[0].getYear();
        else:
            return None;
-            
+
    # Throws GenreException when the tag contains an unrecognized genre format.
    # Note this method returns a eyeD3.Genre object, not a raw string.
    def getGenre(self):
@@ -627,7 +661,7 @@ class Tag:
          elif len(n) == 2:
             tn = self.toInt(n[0])
             tt = self.toInt(n[1])
-      return (tn, tt)      
+      return (tn, tt)
 
    # Returns a tuple with the first value containing the track number and the
    # second the total number of tracks.  One or both of these values may be
@@ -646,10 +680,20 @@ class Tag:
    def getComments(self):
       return self.frames[COMMENT_FID];
 
+   # Since multiple lyrics frames are allowed this returns a list with 0
+   # or more elements.  The elements are not the lyrics strings, they are
+   # eyeD3.frames.LyricsFrame objects.
+   def getLyrics(self):
+      return self.frames[LYRICS_FID];
+
    # Returns a list (possibly empty) of eyeD3.frames.ImageFrame objects.
    def getImages(self):
       return self.frames[IMAGE_FID];
 
+   # Returns a list (possibly empty) of eyeD3.frames.ObjectFrame objects.
+   def getObjects(self):
+      return self.frames[OBJECT_FID];
+
    # Returns a list (possibly empty) of eyeD3.frames.URLFrame objects.
    # Both URLFrame and UserURLFrame objects are returned.  UserURLFrames
    # add a description and encoding, and have a different frame ID.
@@ -765,7 +809,7 @@ class Tag:
 
    def setDiscNum(self, n):
       self.setNum(DISCNUM_FID, n)
-   
+
    def setNum(self, fid, n):
       if n[0] == None and n[1] == None:
          self.frames.removeFramesByID(fid);
@@ -818,6 +862,21 @@ class Tag:
                                      self.strToUnicode(desc),
                                      lang, self.encoding);
 
+   # Add lyrics.  Semantics similar to addComment
+   def addLyrics(self, lyr, desc = u"", lang = DEFAULT_LANG):
+      if not lyr:
+         # A little more than a call to removeFramesByID is involved since we
+         # need to look at more than the frame ID.
+         lyrics = self.frames[LYRICS_FID];
+         for l in lyrics:
+            if l.lang == lang and l.description == desc:
+               self.frames.remove(l);
+               break;
+      else:
+         self.frames.setLyricsFrame(self.strToUnicode(lyr),
+                                     self.strToUnicode(desc),
+                                     lang, self.encoding);
+
    # Semantics similar to addComment
    def addUserTextFrame(self, desc, text):
       if not text:
@@ -833,6 +892,9 @@ class Tag:
    def removeComments(self):
        return self.frames.removeFramesByID(COMMENT_FID);
 
+   def removeLyrics(self):
+       return self.frames.removeFramesByID(LYRICS_FID);
+
    def addImage(self, type, image_file_path, desc = u""):
        if image_file_path:
            image_frame = ImageFrame.create(type, image_file_path, desc);
@@ -844,6 +906,17 @@ class Tag:
                    self.frames.remove(i);
                    break;
 
+   def addObject(self, object_file_path, mime = "", desc = u"",
+                 filename = None ):
+       object_frames = self.frames[OBJECT_FID];
+       for i in object_frames:
+           if i.description == desc:
+               self.frames.remove(i);
+       if object_file_path:
+           object_frame = ObjectFrame.create(object_file_path, mime, desc,
+                                             filename);
+           self.frames.addFrame(object_frame);
+
    def getPlayCount(self):
        if self.frames[PLAYCOUNT_FID]:
            pc = self.frames[PLAYCOUNT_FID][0];
@@ -863,7 +936,7 @@ class Tag:
            frameHeader.id = PLAYCOUNT_FID;
            pc = PlayCountFrame(frameHeader, count = count);
            self.frames.addFrame(pc);
-   
+
    def incrementPlayCount(self, n = 1):
        pc = self.getPlayCount();
        if pc != None:
@@ -917,13 +990,13 @@ class Tag:
       if v != ID3_CURRENT_VERSION:
          self.header.setVersion(v);
          self.frames.setTagHeader(self.header);
-   
+
    def setTextFrame(self, fid, txt):
        if not txt:
           self.frames.removeFramesByID(fid);
        else:
           self.frames.setTextFrame(fid, self.strToUnicode(txt), self.encoding);
-   
+
    def setTextEncoding(self, enc):
        if enc != LATIN1_ENCODING and enc != UTF_16_ENCODING and\
           enc != UTF_16BE_ENCODING and enc != UTF_8_ENCODING:
@@ -996,7 +1069,7 @@ class Tag:
             cmt = cmt[0:28] + "\x00" + chr(int(track) & 0xff);
       tag += cmt;
 
-      if not self.getGenre():
+      if not self.getGenre() or self.getGenre().getId() is None:
          genre = 0;
       else:
          genre = self.getGenre().getId();
@@ -1019,7 +1092,7 @@ class Tag:
       tagFile.write(tag);
       tagFile.flush();
       tagFile.close();
-      
+
    def _fixToWidth(self, s, n):
       retval = str(s);
       retval = retval[0:n];
@@ -1028,7 +1101,7 @@ class Tag:
 
    # Returns false when an ID3 v1 tag is not present, or contains no data.
    def __loadV1Tag(self, f):
-      if isinstance(f, str):
+      if isinstance(f, str) or isinstance(f, unicode):
          fp = file(f, "rb")
          closeFile = 1;
       else:
@@ -1145,6 +1218,7 @@ class Tag:
       rewriteFile = 0;
       paddingSize = 0;
       headerData = "";
+      extHeaderData = "";
 
       # Extended header
       if self.header.extended:
@@ -1154,15 +1228,17 @@ class Tag:
          rewriteFile = 1;
          paddingSize = 2048;
          TRACE_MSG("Rendering extended header");
-         headerData += self.extendedHeader.render(self.header, frameData,
-                                                  paddingSize);
+         extHeaderData += self.extendedHeader.render(self.header, frameData,
+                                                     paddingSize);
 
-      if rewriteFile or (10 + len(headerData) + len(frameData)) >= currTagSize:
+      if rewriteFile or (10 + len(headerData) + len(extHeaderData) +\
+                         len(frameData)) >= currTagSize:
          TRACE_MSG("File rewrite required");
          rewriteFile = 1;
          paddingSize = 2048;
       else:
-         paddingSize = currTagSize - (len(headerData) + len(frameData));
+         paddingSize = currTagSize - (len(headerData) + len(extHeaderData) +\
+                                      len(frameData));
       frameData += ("\x00" * paddingSize);
 
       # Render the tag header.
@@ -1171,7 +1247,7 @@ class Tag:
       headerData = self.header.render(len(frameData));
 
       # Assemble frame.
-      tagData = headerData + frameData;
+      tagData = headerData + extHeaderData + frameData;
 
       # Write the tag.
       if not rewriteFile:
@@ -1180,19 +1256,20 @@ class Tag:
          tagFile.write(tagData);
          tagFile.close();
       else:
-         # Open original
-         tagFile = file(self.linkedFile.name, "rb");
-         # Read all audio data
-         tagFile.seek(currTagSize);
-         audioData = tagFile.read();
-         tagFile.close();
-
+         # FIXME
+         print "POINT #1"
          # Open tmp file
          tmpName = tempfile.mktemp();
          tmpFile = file(tmpName, "w+b");
          TRACE_MSG("Writing %d bytes of tag data" % len(tagData));
          tmpFile.write(tagData);
-         tmpFile.write(audioData);
+
+         # Write audio data in chunks
+         tagFile = file(self.linkedFile.name, "rb");
+         tagFile.seek(currTagSize);
+         self.__copyRemaining(tagFile, tmpFile);
+
+         tagFile.close();
          tmpFile.close();
 
          # Move tmp to orig.
@@ -1209,7 +1286,7 @@ class Tag:
    # Returns >= 0 to indicate the padding size of the read frame; -1 returned
    # when not tag was found.
    def __loadV2Tag(self, f):
-      if isinstance(f, str):
+      if isinstance(f, str) or isinstance(f, unicode):
          fp = file(f, "rb")
          closeFile = 1;
       else:
@@ -1260,6 +1337,18 @@ class Tag:
          raise TagException("Invalid date field: " + fStr);
       return fStr;
 
+   def __copyRemaining(self, src_fp, dest_fp):
+       # Write audio data in chunks
+       done = False
+       amt = 1024 * 512
+       while not done:
+           data = src_fp.read(amt)
+           if data:
+               dest_fp.write(data)
+           else:
+               done = True
+           del data
+
    # DEPRECATED
    # This method will return the first comment in the FrameSet
    # and not all of them.  Multiple COMM frames are common and useful.  Use
@@ -1358,7 +1447,7 @@ class Genre:
          except:
             raise GenreException("eyeD3.genres[" + str(id) + "] " +\
                                  "does not match " + name);
-      
+
    # Parses genre information from genreStr. 
    # The following formats are supported:
    # 01, 2, 23, 125 - ID3 v1 style.
@@ -1399,7 +1488,7 @@ class Genre:
          (id, name) = m.groups();
          if len(id) != 1 and id[0] == '0':
             id = id[1:];
-            
+
          if id and name:
             self.set(int(id), name.strip());
          else:
@@ -1408,7 +1497,7 @@ class Genre:
 
       # Non standard, but witnessed.
       # Match genreName alone.  e.g. Rap, Rock, blues.
-      regex = re.compile("^[A-Z 0-9+/\-&]+\00*$", re.IGNORECASE);
+      regex = re.compile("^[A-Z 0-9'+/\-&]+\00*$", re.IGNORECASE);
       if regex.match(genreStr):
          self.setName(genreStr);
          return;
@@ -1446,13 +1535,13 @@ class TagFile:
          self.fileSize = os.stat(self.fileName)[ST_SIZE];
       return self.fileSize;
 
-   def rename(self, name):
+   def rename(self, name, fsencoding):
        base = os.path.basename(self.fileName);
        base_ext = os.path.splitext(base)[1];
        dir = os.path.dirname(self.fileName);
        if not dir:
            dir = ".";
-       new_name = dir + os.sep + name.encode("iso8859-1") + base_ext;
+       new_name = dir + os.sep + name.encode(fsencoding) + base_ext;
        try:
            os.rename(self.fileName, new_name);
            self.fileName = new_name;
@@ -1488,7 +1577,7 @@ class Mp3AudioFile(TagFile):
          raise self.invalidFileExc;
 
       # Parse ID3 tag.
-      f = file(fileName, "rb");
+      f = file(self.fileName, "rb");
       tag = Tag();
       hasTag = tag.link(f, tagVersion);
       # Find the first mp3 frame.
@@ -1501,7 +1590,6 @@ class Mp3AudioFile(TagFile):
          # XXX: Note that v2.4 allows for appended tags; account for that.
          framePos = tag.header.SIZE + tag.header.tagSize;
 
-      # XXX: This algorithm is slow, if the file is long it could take awhile.
       f.seek(framePos);
       bString = f.read(4);
       if len(bString) < 4:
@@ -1509,14 +1597,30 @@ class Mp3AudioFile(TagFile):
                                            "frame");
       frameHead = bin2dec(bytes2bin(bString));
       header = mp3.Header();
+      it_count = 0;
       # Keep reading until we find a valid mp3 frame header.
       while not header.isValid(frameHead):
+         # Originally, the search was one byte at a time.  Occasionally a tag
+         # was heavily, and incorrectly, padded or a corrupt mp3 would cause
+         # a byte by byte search to take a long time.  This algorithm speeds
+         # up this particular case.
+         if it_count > 9:
+             f.seek(f.tell() + 128);
+             bString = f.read(4);
+             if len(bString) < 4:
+                raise InvalidAudioFormatException("Unable to find a valid mp3 "\
+                                                  "frame");
+             frameHead = bin2dec(bytes2bin(bString));
+             it_count = 0;
+             continue;
+
          frameHead <<= 8;
          bString = f.read(1);
          if len(bString) != 1:
             raise InvalidAudioFormatException("Unable to find a valid mp3 "\
                                               "frame");
          frameHead |= ord(bString[0]);
+         it_count += 1;
       TRACE_MSG("mp3 header %x found at position: %d (0x%x)" % \
                 (frameHead, f.tell() - 4, f.tell() - 4));
 
@@ -1550,7 +1654,7 @@ class Mp3AudioFile(TagFile):
                length -= 128;
          elif tag and tag.isV1():
             length -= 128;
-         self.play_time = int((length / header.frameLength) * tpf);    
+         self.play_time = int((length / header.frameLength) * tpf);
 
       self.header = header;
       self.xingHeader = xingHeader;
@@ -1780,7 +1884,7 @@ class GenreMap(list):
       while count < 256:
          self.append("Unknown");
          count += 1;
-      
+
       for index in range(len(self)):
          if self[index]:
             self.reverseDict[string.lower(self[index])] = index
@@ -1790,25 +1894,35 @@ class LinkedFile:
    tagSize = 0;  # This includes the padding byte count.
 
    def __init__(self, fileName):
-      self.name = fileName;
+       if isinstance(fileName, str):
+           try:
+               self.name = unicode(fileName, sys.getfilesystemencoding());
+           except:
+               # Work around the local encoding not matching that of a mounted
+               # filesystem
+               self.name = fileName
+       else:
+           self.name = fileName;
 
 def tagToUserTune(tag):
     audio_file = None;
     if isinstance(tag, Mp3AudioFile):
         audio_file = tag;
         tag = audio_file.getTag();
-        
-    tune =  "<tune xmlns='http://jabber.org/protocol/tune'>\n";
+
+    tune =  u"<tune xmlns='http://jabber.org/protocol/tune'>\n";
     if tag.getArtist():
         tune += "  <artist>" + tag.getArtist() + "</artist>\n";
     if tag.getTitle():
         tune += "  <title>" + tag.getTitle() + "</title>\n";
     if tag.getAlbum():
         tune += "  <source>" + tag.getAlbum() + "</source>\n";
-    tune += "  <track>" + "file://" + os.path.abspath(tag.linkedFile.name) +\
+    tune += "  <track>" +\
+            "file://" + unicode(os.path.abspath(tag.linkedFile.name)) +\
             "</track>\n";
     if audio_file:
-        tune += "  <length>" + str(audio_file.getPlayTime()) + "</length>\n";
+        tune += "  <length>" + unicode(audio_file.getPlayTime()) +\
+                "</length>\n";
     tune += "</tune>\n";
     return tune;
 
